NODE *getInheritedProcWithParent(NODE *name, NODE *obj, NODE **parent);
NODE *getInheritedProcWithParentList(NODE *name,
                                     NODE *parentList, NODE **parent);
NODE *getInheritedProc(NODE *name, NODE *obj);
