/*
**  This file was automatically generated from gpl.html.  Make changes
**  to that file rather than this one.
*/

const char *gplText =
  _T("<HTML>\n")
  _T("<HEAD>\n")
  _T("<TITLE>GNU General Public License</TITLE>\n")
  _T("</HEAD>\n")
  _T("<BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\" LINK=\"#1F00FF\" ALINK=\"#FF0000\" VLINK=\"#9900DD\">\n")
  _T("<H2>GNU GENERAL PUBLIC LICENSE</H2>\n")
  _T("<P>\n")
  _T("Version 2, June 1991\n")
  _T("\n")
  _T("</P>\n")
  _T("\n")
  _T("<PRE>\n")
  _T("Copyright (C) 1989, 1991 Free Software Foundation, Inc.\n")
  _T("59 Temple Place - Suite 330, Boston, MA  02111-1307, USA\n")
  _T("\n")
  _T("Everyone is permitted to copy and distribute verbatim copies\n")
  _T("of this license document, but changing it is not allowed.\n")
  _T("</PRE>\n")
  _T("\n")
  _T("\n")
  _T("\n")
  _T("<H2>Preamble</H2>\n")
  _T("\n")
  _T("<P>\n")
  _T("  The licenses for most software are designed to take away your\n")
  _T("freedom to share and change it.  By contrast, the GNU General Public\n")
  _T("License is intended to guarantee your freedom to share and change free\n")
  _T("software--to make sure the software is free for all its users.  This\n")
  _T("General Public License applies to most of the Free Software\n")
  _T("Foundation's software and to any other program whose authors commit to\n")
  _T("using it.  (Some other Free Software Foundation software is covered by\n")
  _T("the GNU Library General Public License instead.)  You can apply it to\n")
  _T("your programs, too.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  When we speak of free software, we are referring to freedom, not\n")
  _T("price.  Our General Public Licenses are designed to make sure that you\n")
  _T("have the freedom to distribute copies of free software (and charge for\n")
  _T("this service if you wish), that you receive source code or can get it\n")
  _T("if you want it, that you can change the software or use pieces of it\n")
  _T("in new free programs; and that you know you can do these things.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  To protect your rights, we need to make restrictions that forbid\n")
  _T("anyone to deny you these rights or to ask you to surrender the rights.\n")
  _T("These restrictions translate to certain responsibilities for you if you\n")
  _T("distribute copies of the software, or if you modify it.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  For example, if you distribute copies of such a program, whether\n")
  _T("gratis or for a fee, you must give the recipients all the rights that\n")
  _T("you have.  You must make sure that they, too, receive or can get the\n")
  _T("source code.  And you must show them these terms so they know their\n")
  _T("rights.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  We protect your rights with two steps: (1) copyright the software, and\n")
  _T("(2) offer you this license which gives you legal permission to copy,\n")
  _T("distribute and/or modify the software.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  Also, for each author's protection and ours, we want to make certain\n")
  _T("that everyone understands that there is no warranty for this free\n")
  _T("software.  If the software is modified by someone else and passed on, we\n")
  _T("want its recipients to know that what they have is not the original, so\n")
  _T("that any problems introduced by others will not reflect on the original\n")
  _T("authors' reputations.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  Finally, any free program is threatened constantly by software\n")
  _T("patents.  We wish to avoid the danger that redistributors of a free\n")
  _T("program will individually obtain patent licenses, in effect making the\n")
  _T("program proprietary.  To prevent this, we have made it clear that any\n")
  _T("patent must be licensed for everyone's free use or not licensed at all.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  The precise terms and conditions for copying, distribution and\n")
  _T("modification follow.\n")
  _T("\n")
  _T("</P>\n")
  _T("\n")
  _T("\n")
  _T("<H2>TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION</H2>\n")
  _T("\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>0.</STRONG>\n")
  _T(" This License applies to any program or other work which contains\n")
  _T("a notice placed by the copyright holder saying it may be distributed\n")
  _T("under the terms of this General Public License.  The \"Program\", below,\n")
  _T("refers to any such program or work, and a \"work based on the Program\"\n")
  _T("means either the Program or any derivative work under copyright law:\n")
  _T("that is to say, a work containing the Program or a portion of it,\n")
  _T("either verbatim or with modifications and/or translated into another\n")
  _T("language.  (Hereinafter, translation is included without limitation in\n")
  _T("the term \"modification\".)  Each licensee is addressed as \"you\".\n")
  _T("<P>\n")
  _T("\n")
  _T("Activities other than copying, distribution and modification are not\n")
  _T("covered by this License; they are outside its scope.  The act of\n")
  _T("running the Program is not restricted, and the output from the Program\n")
  _T("is covered only if its contents constitute a work based on the\n")
  _T("Program (independent of having been made by running the Program).\n")
  _T("Whether that is true depends on what the Program does.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>1.</STRONG>\n")
  _T(" You may copy and distribute verbatim copies of the Program's\n")
  _T("source code as you receive it, in any medium, provided that you\n")
  _T("conspicuously and appropriately publish on each copy an appropriate\n")
  _T("copyright notice and disclaimer of warranty; keep intact all the\n")
  _T("notices that refer to this License and to the absence of any warranty;\n")
  _T("and give any other recipients of the Program a copy of this License\n")
  _T("along with the Program.\n")
  _T("<P>\n")
  _T("\n")
  _T("You may charge a fee for the physical act of transferring a copy, and\n")
  _T("you may at your option offer warranty protection in exchange for a fee.\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>2.</STRONG>\n")
  _T(" You may modify your copy or copies of the Program or any portion\n")
  _T("of it, thus forming a work based on the Program, and copy and\n")
  _T("distribute such modifications or work under the terms of Section 1\n")
  _T("above, provided that you also meet all of these conditions:\n")
  _T("<P>\n")
  _T("\n")
  _T("<UL>\n")
  _T("\n")
  _T("<LI><STRONG>a)</STRONG>\n")
  _T("     You must cause the modified files to carry prominent notices\n")
  _T("     stating that you changed the files and the date of any change.\n")
  _T("\n")
  _T("<P>\n")
  _T("<LI><STRONG>b)</STRONG>\n")
  _T("     You must cause any work that you distribute or publish, that in\n")
  _T("     whole or in part contains or is derived from the Program or any\n")
  _T("     part thereof, to be licensed as a whole at no charge to all third\n")
  _T("     parties under the terms of this License.\n")
  _T("\n")
  _T("<P>\n")
  _T("<LI><STRONG>c)</STRONG>\n")
  _T("     If the modified program normally reads commands interactively\n")
  _T("     when run, you must cause it, when started running for such\n")
  _T("     interactive use in the most ordinary way, to print or display an\n")
  _T("     announcement including an appropriate copyright notice and a\n")
  _T("     notice that there is no warranty (or else, saying that you provide\n")
  _T("     a warranty) and that users may redistribute the program under\n")
  _T("     these conditions, and telling the user how to view a copy of this\n")
  _T("     License.  (Exception: if the Program itself is interactive but\n")
  _T("     does not normally print such an announcement, your work based on\n")
  _T("     the Program is not required to print an announcement.)\n")
  _T("</UL>\n")
  _T("<p>\n")
  _T("\n")
  _T("These requirements apply to the modified work as a whole.  If\n")
  _T("identifiable sections of that work are not derived from the Program,\n")
  _T("and can be reasonably considered independent and separate works in\n")
  _T("themselves, then this License, and its terms, do not apply to those\n")
  _T("sections when you distribute them as separate works.  But when you\n")
  _T("distribute the same sections as part of a whole which is a work based\n")
  _T("on the Program, the distribution of the whole must be on the terms of\n")
  _T("this License, whose permissions for other licensees extend to the\n")
  _T("entire whole, and thus to each and every part regardless of who wrote it.\n")
  _T("<P>\n")
  _T("\n")
  _T("Thus, it is not the intent of this section to claim rights or contest\n")
  _T("your rights to work written entirely by you; rather, the intent is to\n")
  _T("exercise the right to control the distribution of derivative or\n")
  _T("collective works based on the Program.\n")
  _T("<P>\n")
  _T("\n")
  _T("In addition, mere aggregation of another work not based on the Program\n")
  _T("with the Program (or with a work based on the Program) on a volume of\n")
  _T("a storage or distribution medium does not bring the other work under\n")
  _T("the scope of this License.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>3.</STRONG>\n")
  _T(" You may copy and distribute the Program (or a work based on it,\n")
  _T("under Section 2) in object code or executable form under the terms of\n")
  _T("Sections 1 and 2 above provided that you also do one of the following:\n")
  _T("\n")
  _T("\n")
  _T("<!-- we use this doubled UL to get the sub-sections indented, -->\n")
  _T("<!-- while making the bullets as unobvious as possible. -->\n")
  _T("<UL>\n")
  _T("\n")
  _T("<LI><STRONG>a)</STRONG>\n")
  _T("     Accompany it with the complete corresponding machine-readable\n")
  _T("     source code, which must be distributed under the terms of Sections\n")
  _T("     1 and 2 above on a medium customarily used for software interchange; or,\n")
  _T("\n")
  _T("<P>\n")
  _T("<LI><STRONG>b)</STRONG>\n")
  _T("     Accompany it with a written offer, valid for at least three\n")
  _T("     years, to give any third party, for a charge no more than your\n")
  _T("     cost of physically performing source distribution, a complete\n")
  _T("     machine-readable copy of the corresponding source code, to be\n")
  _T("     distributed under the terms of Sections 1 and 2 above on a medium\n")
  _T("     customarily used for software interchange; or,\n")
  _T("\n")
  _T("<P>\n")
  _T("<LI><STRONG>c)</STRONG>\n")
  _T("     Accompany it with the information you received as to the offer\n")
  _T("     to distribute corresponding source code.  (This alternative is\n")
  _T("     allowed only for noncommercial distribution and only if you\n")
  _T("     received the program in object code or executable form with such\n")
  _T("     an offer, in accord with Subsection b above.)\n")
  _T("</UL>\n")
  _T("<p>\n")
  _T("\n")
  _T("The source code for a work means the preferred form of the work for\n")
  _T("making modifications to it.  For an executable work, complete source\n")
  _T("code means all the source code for all modules it contains, plus any\n")
  _T("associated interface definition files, plus the scripts used to\n")
  _T("control compilation and installation of the executable.  However, as a\n")
  _T("special exception, the source code distributed need not include\n")
  _T("anything that is normally distributed (in either source or binary\n")
  _T("form) with the major components (compiler, kernel, and so on) of the\n")
  _T("operating system on which the executable runs, unless that component\n")
  _T("itself accompanies the executable.\n")
  _T("<P>\n")
  _T("\n")
  _T("If distribution of executable or object code is made by offering\n")
  _T("access to copy from a designated place, then offering equivalent\n")
  _T("access to copy the source code from the same place counts as\n")
  _T("distribution of the source code, even though third parties are not\n")
  _T("compelled to copy the source along with the object code.\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>4.</STRONG>\n")
  _T(" You may not copy, modify, sublicense, or distribute the Program\n")
  _T("except as expressly provided under this License.  Any attempt\n")
  _T("otherwise to copy, modify, sublicense or distribute the Program is\n")
  _T("void, and will automatically terminate your rights under this License.\n")
  _T("However, parties who have received copies, or rights, from you under\n")
  _T("this License will not have their licenses terminated so long as such\n")
  _T("parties remain in full compliance.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>5.</STRONG>\n")
  _T(" You are not required to accept this License, since you have not\n")
  _T("signed it.  However, nothing else grants you permission to modify or\n")
  _T("distribute the Program or its derivative works.  These actions are\n")
  _T("prohibited by law if you do not accept this License.  Therefore, by\n")
  _T("modifying or distributing the Program (or any work based on the\n")
  _T("Program), you indicate your acceptance of this License to do so, and\n")
  _T("all its terms and conditions for copying, distributing or modifying\n")
  _T("the Program or works based on it.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>6.</STRONG>\n")
  _T(" Each time you redistribute the Program (or any work based on the\n")
  _T("Program), the recipient automatically receives a license from the\n")
  _T("original licensor to copy, distribute or modify the Program subject to\n")
  _T("these terms and conditions.  You may not impose any further\n")
  _T("restrictions on the recipients' exercise of the rights granted herein.\n")
  _T("You are not responsible for enforcing compliance by third parties to\n")
  _T("this License.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>7.</STRONG>\n")
  _T(" If, as a consequence of a court judgment or allegation of patent\n")
  _T("infringement or for any other reason (not limited to patent issues),\n")
  _T("conditions are imposed on you (whether by court order, agreement or\n")
  _T("otherwise) that contradict the conditions of this License, they do not\n")
  _T("excuse you from the conditions of this License.  If you cannot\n")
  _T("distribute so as to satisfy simultaneously your obligations under this\n")
  _T("License and any other pertinent obligations, then as a consequence you\n")
  _T("may not distribute the Program at all.  For example, if a patent\n")
  _T("license would not permit royalty-free redistribution of the Program by\n")
  _T("all those who receive copies directly or indirectly through you, then\n")
  _T("the only way you could satisfy both it and this License would be to\n")
  _T("refrain entirely from distribution of the Program.\n")
  _T("<P>\n")
  _T("\n")
  _T("If any portion of this section is held invalid or unenforceable under\n")
  _T("any particular circumstance, the balance of the section is intended to\n")
  _T("apply and the section as a whole is intended to apply in other\n")
  _T("circumstances.\n")
  _T("<P>\n")
  _T("\n")
  _T("It is not the purpose of this section to induce you to infringe any\n")
  _T("patents or other property right claims or to contest validity of any\n")
  _T("such claims; this section has the sole purpose of protecting the\n")
  _T("integrity of the free software distribution system, which is\n")
  _T("implemented by public license practices.  Many people have made\n")
  _T("generous contributions to the wide range of software distributed\n")
  _T("through that system in reliance on consistent application of that\n")
  _T("system; it is up to the author/donor to decide if he or she is willing\n")
  _T("to distribute software through any other system and a licensee cannot\n")
  _T("impose that choice.\n")
  _T("<P>\n")
  _T("\n")
  _T("This section is intended to make thoroughly clear what is believed to\n")
  _T("be a consequence of the rest of this License.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>8.</STRONG>\n")
  _T(" If the distribution and/or use of the Program is restricted in\n")
  _T("certain countries either by patents or by copyrighted interfaces, the\n")
  _T("original copyright holder who places the Program under this License\n")
  _T("may add an explicit geographical distribution limitation excluding\n")
  _T("those countries, so that distribution is permitted only in or among\n")
  _T("countries not thus excluded.  In such case, this License incorporates\n")
  _T("the limitation as if written in the body of this License.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>9.</STRONG>\n")
  _T(" The Free Software Foundation may publish revised and/or new versions\n")
  _T("of the General Public License from time to time.  Such new versions will\n")
  _T("be similar in spirit to the present version, but may differ in detail to\n")
  _T("address new problems or concerns.\n")
  _T("<P>\n")
  _T("\n")
  _T("Each version is given a distinguishing version number.  If the Program\n")
  _T("specifies a version number of this License which applies to it and \"any\n")
  _T("later version\", you have the option of following the terms and conditions\n")
  _T("either of that version or of any later version published by the Free\n")
  _T("Software Foundation.  If the Program does not specify a version number of\n")
  _T("this License, you may choose any version ever published by the Free Software\n")
  _T("Foundation.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("\n")
  _T("<STRONG>10.</STRONG>\n")
  _T(" If you wish to incorporate parts of the Program into other free\n")
  _T("programs whose distribution conditions are different, write to the author\n")
  _T("to ask for permission.  For software which is copyrighted by the Free\n")
  _T("Software Foundation, write to the Free Software Foundation; we sometimes\n")
  _T("make exceptions for this.  Our decision will be guided by the two goals\n")
  _T("of preserving the free status of all derivatives of our free software and\n")
  _T("of promoting the sharing and reuse of software generally.\n")
  _T("\n")
  _T("\n")
  _T("\n")
  _T("<P><STRONG>NO WARRANTY</STRONG></P>\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>11.</STRONG>\n")
  _T(" BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n")
  _T("FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n")
  _T("OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n")
  _T("PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n")
  _T("OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n")
  _T("MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n")
  _T("TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n")
  _T("PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n")
  _T("REPAIR OR CORRECTION.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("<STRONG>12.</STRONG>\n")
  _T(" IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n")
  _T("WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n")
  _T("REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n")
  _T("INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n")
  _T("OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n")
  _T("TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n")
  _T("YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n")
  _T("PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n")
  _T("POSSIBILITY OF SUCH DAMAGES.\n")
  _T("\n")
  _T("<P>\n")
  _T("\n")
  _T("\n")
  _T("<H2>END OF TERMS AND CONDITIONS</H2>\n")
  _T("\n")
  _T("\n")
  _T("\n")
  _T("<H2>How to Apply These Terms to Your New Programs</H2>\n")
  _T("\n")
  _T("<P>\n")
  _T("  If you develop a new program, and you want it to be of the greatest\n")
  _T("possible use to the public, the best way to achieve this is to make it\n")
  _T("free software which everyone can redistribute and change under these terms.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("  To do so, attach the following notices to the program.  It is safest\n")
  _T("to attach them to the start of each source file to most effectively\n")
  _T("convey the exclusion of warranty; and each file should have at least\n")
  _T("the \"copyright\" line and a pointer to where the full notice is found.\n")
  _T("\n")
  _T("</P>\n")
  _T("\n")
  _T("<PRE>\n")
  _T("<VAR>one line to give the program's name and an idea of what it does.</VAR>\n")
  _T("Copyright (C) <VAR>yyyy</VAR>  <VAR>name of author</VAR>\n")
  _T("\n")
  _T("This program is free software; you can redistribute it and/or\n")
  _T("modify it under the terms of the GNU General Public License\n")
  _T("as published by the Free Software Foundation; either version 2\n")
  _T("of the License, or (at your option) any later version.\n")
  _T("\n")
  _T("This program is distributed in the hope that it will be useful,\n")
  _T("but WITHOUT ANY WARRANTY; without even the implied warranty of\n")
  _T("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n")
  _T("GNU General Public License for more details.\n")
  _T("\n")
  _T("You should have received a copy of the GNU General Public License\n")
  _T("along with this program; if not, write to the Free Software\n")
  _T("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n")
  _T("</PRE>\n")
  _T("\n")
  _T("<P>\n")
  _T("Also add information on how to contact you by electronic and paper mail.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("If the program is interactive, make it output a short notice like this\n")
  _T("when it starts in an interactive mode:\n")
  _T("\n")
  _T("</P>\n")
  _T("\n")
  _T("<PRE>\n")
  _T("Gnomovision version 69, Copyright (C) <VAR>year</VAR> <VAR>name of author</VAR>\n")
  _T("Gnomovision comes with ABSOLUTELY NO WARRANTY; for details\n")
  _T("type `show w'.  This is free software, and you are welcome\n")
  _T("to redistribute it under certain conditions; type `show c'\n")
  _T("for details.\n")
  _T("</PRE>\n")
  _T("\n")
  _T("<P>\n")
  _T("The hypothetical commands <SAMP>`show w'</SAMP> and <SAMP>`show c'</SAMP> should show\n")
  _T("the appropriate parts of the General Public License.  Of course, the\n")
  _T("commands you use may be called something other than <SAMP>`show w'</SAMP> and\n")
  _T("<SAMP>`show c'</SAMP>; they could even be mouse-clicks or menu items--whatever\n")
  _T("suits your program.\n")
  _T("\n")
  _T("</P>\n")
  _T("<P>\n")
  _T("You should also get your employer (if you work as a programmer) or your\n")
  _T("school, if any, to sign a \"copyright disclaimer\" for the program, if\n")
  _T("necessary.  Here is a sample; alter the names:\n")
  _T("\n")
  _T("</P>\n")
  _T("\n")
  _T("<PRE>\n")
  _T("Yoyodyne, Inc., hereby disclaims all copyright\n")
  _T("interest in the program `Gnomovision'\n")
  _T("(which makes passes at compilers) written\n")
  _T("by James Hacker.\n")
  _T("\n")
  _T("<VAR>signature of Ty Coon</VAR>, 1 April 1989\n")
  _T("Ty Coon, President of Vice\n")
  _T("</PRE>\n")
  _T("\n")
  _T("<P>\n")
  _T("This General Public License does not permit incorporating your program into\n")
  _T("proprietary programs.  If your program is a subroutine library, you may\n")
  _T("consider it more useful to permit linking proprietary applications with the\n")
  _T("library.  If this is what you want to do, use the GNU Library General\n")
  _T("Public License instead of this License.\n")
  _T("\n")
  _T("</BODY>\n")
  _T("</HTML>\n")
;
