char *libloc=LIBLOC;		/* "/usr/local/lib/logo/logolib" */
char *helploc=HELPLOC;		/* "/usr/local/lib/logo/helpfiles" */
char *cslsloc=CSLSLOC;		/* "/usr/local/lib/logo/csls" */
char *temploc=TEMPLOC;		/* "/tmp" */
char *separator="/";
